//
// Copyright (c) 2022, MindFusion LLC - Bulgaria.
//

using System;
using System.Drawing;
using System.Drawing.Imaging;

using MindFusion.Drawing;
using MindFusion.Diagramming.WinForms.Behaviors;


namespace MindFusion.Diagramming.WinForms.Samples.CS.Reflections
{
	class ReflectiveNode : ShapeNode
	{
		public ReflectiveNode()
		{
			reflectionDistance = 2;
			reflectionOpacity = 0.75f;
			reflectionStretch = 1;
		}

		public ReflectiveNode(Diagram diagram)
			: base(diagram)
		{
			reflectionDistance = 2;
			reflectionOpacity = 0.75f;
			reflectionStretch = 1;
		}

		public override void DrawLocal(IGraphics graphics, RenderOptions options)
		{
			// Reflections work only with GDI drawing targets
			GdiGraphics gdiGraphics = graphics as GdiGraphics;
			if (gdiGraphics == null)
				return;

			RectangleF bounds = GetLocalBounds();
			if (bounds.Width < 1 || bounds.Height < 1)
				return;

			// Calculate the node size in pixels
			float unitsPerPixel = Constants.GetPixel(MeasureUnit);
			if (unitsPerPixel <= float.Epsilon)
				return;

			// Bitmap size in pixels (with 1px tolerance on all sides to
			// accommodate for antialiasing)
			int width = (int)Math.Ceiling((bounds.Width + 2 * unitsPerPixel) / unitsPerPixel);
			int height = (int)Math.Ceiling((bounds.Height + 2 * unitsPerPixel) / unitsPerPixel);

			Bitmap offscreen = new Bitmap(width, height, PixelFormat.Format32bppArgb);
			Graphics g = Graphics.FromImage(offscreen);
			g.TextRenderingHint = System.Drawing.Text.TextRenderingHint.AntiAliasGridFit;
			g.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.AntiAlias;

			// Make sure we are drawing at (1, 1)
			g.TranslateTransform(unitsPerPixel, unitsPerPixel);
			g.PageUnit = graphics.PageUnit;
				
			base.DrawLocal(new GdiGraphics(g), options);

			graphics.DrawImage(offscreen, bounds.X - unitsPerPixel, bounds.Y - unitsPerPixel);

			// Draw reflection by inverting the y-axis of the graphics
			graphics.TranslateTransform(0, bounds.Height);
			graphics.ScaleTransform(1, -1);
			graphics.TranslateTransform(0, -bounds.Height);

			AddOpacity(offscreen);

			graphics.DrawImage(offscreen, bounds.X - unitsPerPixel, bounds.Y + unitsPerPixel - reflectionDistance);

			g.Dispose();
			offscreen.Dispose();
		}

		public override void DrawShadowLocal(IGraphics graphics, RenderOptions options)
		{
			// Do not draw shadows for reflection nodes
			//base.DrawShadowLocal(graphics, options);
		}

		public override RectangleF GetRepaintRect(bool includeConnected)
		{
			RectangleF rect = base.GetRepaintRect(includeConnected);
			rect.Height = rect.Height * (1 + reflectionStretch) + reflectionDistance;
			return rect;
		}

		private void AddOpacity(Bitmap image)
		{
			int w = image.Width;
			int h = image.Height;
			Rectangle rc = new Rectangle(0, 0, w, h);
			BitmapData data = image.LockBits(rc, ImageLockMode.ReadOnly, PixelFormat.Format32bppArgb);
			int stride = data.Stride / 4;

			unsafe
			{
				uint* pixels = (uint*)data.Scan0;
				for (int i = 0; i < w; i++)
				{
					for (int j = 0; j < h; j++)
					{
						// Opacity function f(j):
						// f(0) = 0
						// f(h * k) = 0, for k in [0..1]
						// f(h) = reflectionOpacity
						uint pixel = pixels[j * stride + i];
						float a = (float)(byte)(pixel >> 24);
						float k = (1 - reflectionStretch);
						float f = Sqr(j, h * k, h);
						a = a * (f * reflectionOpacity);
						if (a < 0)
							a = 0;
						if (a > 255)
							a = 255;
						pixels[j * stride + i] = (pixel & 0x00FFFFFF) | ((uint)(byte)a << 24);
					}
				}
			}

			image.UnlockBits(data);
		}

		private float Linear(float x, float min, float max)
		{
			if (min == max)
				return 0;
			if (x < min)
				x = min;
			if (x > max)
				x = max;
			return (x - min) / (max - min);
		}

		private float Sqr(float x, float min, float max)
		{
			float i = Linear(x, min, max);
			return i * i;
		}


		/// <summary>
		/// Gets or sets the distance between the reflection and the node.
		/// </summary>
		public float ReflectionDistance
		{
			get { return reflectionDistance; }
			set { reflectionDistance = value; }
		}

		/// <summary>
		/// Gets or sets the reflection opacity, where it is most visible.
		/// The values range in the interval [0..1] where 0 indicates full
		/// transparency and 1 indicates opaque.
		/// </summary>
		public float ReflectionOpacity
		{
			get { return reflectionOpacity; }
			set { reflectionOpacity = value; }
		}

		/// <summary>
		/// Gets the how much of the node the reflection covers. The values
		/// range in the interval [0..1] where 0 indicates no reflection and
		/// 1 indicates that the reflection covers the full height of the node.
		/// </summary>
		public float ReflectionStretch
		{
			get { return reflectionStretch; }
			set { reflectionStretch = value; }
		}


		private float reflectionDistance;
		private float reflectionOpacity;
		private float reflectionStretch;
	}

	class CustomBehavior : LinkShapesBehavior
	{
		public CustomBehavior(DiagramView diagramView)
			: base(diagramView)
		{
		}

		protected override DiagramNode CreateNode()
		{
			return new ReflectiveNode();
		}
	}
}
